#ifdef INFO_ONLY
/* User modifiable API to determine special keys for Terminal i/o 
 * mode in FlagShip via InkeyStatus() for Unix/Linux. This API 
 * function is not used in GUI mode nor in FlagShip for MS-Windows.
 *
 * If you wish to report the Shift, Ctrl, NumLock etc. keys
 * via InkeyStatus() also in Terminal i/o mode running in X11, 
 * compile this source by
 *   FlagShip -c inkeystatapi.c 
 * and add the object + X11 library to your application
 *   FlagShip myapp*.prg .... inkeystatapi.o -lX11
 * The standard function available in FlagShip library is compiled with
 *   FlagShip -c -DNO_X11_TERM_API inkeystatapi.c
 * i.e. only for plain console w/o X11, to avoid linking the X11 library, 
 * which is often not available on servers or basic Unix/Linux systems.
 *
 * Release $Date: 2005/10/29 10:37:30 $
 */
#endif
  
#ifndef NO_X11_TERM_API
#define USE_X11_TERM_API
#endif

// #define CONSIDER_DISPLAY_CHANGE  

/* *************************** MS-Windows ***********************
 *
 * Do not use this source for FlagShip/Windows, since internal 
 * MS-Windows API functions are used here. Just in case:
 */
#ifdef FS_WIN32
int InkeyStatusApi(mode)
{
   return 0;
}
#else

/* *************************** Unix/Linux ***********************
 */
#include <stdio.h>
#include <stdlib.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <sys/types.h>
#include <linux/kd.h>

static int keyStateX11(int mode);
static int keyStateCons(int mode);

#ifdef USE_X11_TERM_API
/* *************************** Unix/Linux: X11 available ********
 */
#include <X11/X.h>
#include <X11/Xlib.h>
#include <X11/keysym.h>

static int keyStateX11(int mode)
{
   int             ii, iMode, iKeyMask = 0;
   Window          wDummy1, wDummy2;
   int             iDummy3, iDummy4, iDummy5, iDummy6;
   unsigned int    iMask;
   static Display  *pDisplay = 0;
   XModifierKeymap *map;
   KeyCode         keyCode;
   char            *cEnv = 0;
   static char     *cDisplayEnv = 0;
   
   /* XOpenDisplay() is expensive, we therefore open it only once.
    * But if you will change the DISPLAY environment variable during the
    * program execution, enable the #define CONSIDER_DISPLAY_CHANGE above.
    */
#ifdef CONSIDER_DISPLAY_CHANGE  
   cEnv = getenv("DISPLAY");
   if(!cEnv || !*cEnv)
      return -1;                                // envir.var DISPLAY n/a
   if(!cDisplayEnv) {                           // first invocation
      cDisplayEnv = malloc(sizeof(char) * (strlen(cEnv) + 1));
      strcpy(cDisplayEnv, cEnv);
      pDisplay = 0;
   } else if(strcmp(cDisplayEnv, cEnv) != 0) {  // DISPLAY changed
      cDisplayEnv = realloc(cDisplayEnv, sizeof(char) * (strlen(cEnv) + 1));
      strcpy(cDisplayEnv, cEnv);
      if(pDisplay)
         XCloseDisplay(pDisplay);
      pDisplay = 0;                             // needs reinit
   } 
#endif
   if(!pDisplay) {                              // not initialized yet
      if(!cEnv) {                               // check envir.var DISPLAY
         cEnv = getenv("DISPLAY");
         if(!cEnv || !*cEnv)
            return -1;                          // DISPLAY n/a or is empty
      }
      pDisplay = XOpenDisplay(NULL);            // use envir.var DISPLAY 
   }      
   if(!pDisplay) {
//    fprintf(stderr, "-- X11 is not running or is not accessible --\n");
      return -1;
   }
   
   /* ok, X11 is available
    */
   switch(mode) {
   case 1:
      iMode = XK_Scroll_Lock;
      break;
   case 2:
      iMode = XK_Caps_Lock;
      break;
   case 3:
      iMode = XK_Num_Lock;
      break;
   default:
      return -1;
   }      
   map = XGetModifierMapping(pDisplay);
   if(!map)
      return -1;
      
   keyCode = XKeysymToKeycode(pDisplay,iMode);
   if(keyCode == NoSymbol)
      return -1;
      
   for(ii = 0; ii < 8; ++ii) {
      if( map->modifiermap[map->max_keypermod * ii] == keyCode) {
         iKeyMask = 1 << ii;
     }
   }
   XQueryPointer(pDisplay, DefaultRootWindow(pDisplay), &wDummy1, &wDummy2,
                 &iDummy3, &iDummy4, &iDummy5, &iDummy6, &iMask );
   XFreeModifiermap(map);
   return (iMask & iKeyMask) != 0 ? 1 : 0;
}
#else /* ifdef USE_X11_TERM_API */
/* *************************** Unix/Linux: X11 not available ****
 * this is the default in libFlagShip*.so|.a
 */
static int keyStateX11(int mode)
{
   return -1;
}
#endif

/* ************************** Unix/Linux: Console ***************
 * mode = 1: ScrollLock
 *        2: CapsLock
 *        3: NumLock
 *        4: Shift
 * see also: man 4 console_ioctl
 */
static int keyStateCons(int mode)
{
   static int fd = -2, failedShift = 0, failedLock = 0;
   char cc = 0;
   int ret = -1;
   
   if(fd == -2)            // first call
      fd = fileno(stdin);  
   if(fd < 0)
      return -1;  
      
   if(mode == 4) {         // shift
      if(failedShift > 2)
         return -1;        // don't try anew, ioctl is expensive

      cc = 6;
      ret = ioctl(fd, TIOCLINUX, &cc); // TIOCLINUX = 0x541C in asm/ioctls.h
      if(ret == -1)
         failedShift++;
      else
         failedShift = 0;
      return ret;
   }

   if(failedLock > 6)
      return -1;                       // don't try anew, ioctl is expensive
      
   ret = ioctl(fd, KDGKBLED, &cc);     // KDGKBLED = 0x4B64 in linux/kd.h
   if(ret == -1) {
//    fprintf(stderr, "-- ioctl() error, console n/a, Xterm mode? --\n");
      failedLock++;
      return ret; 
   }
   failedLock = 0;
   
   switch(mode) {
   case 1: 
      ret = (cc & 0x01) ? 1 : 0;
      break;
   case 2: 
      ret = (cc & 0x04) ? 1 : 0;
      break;
   case 3: 
      ret = (cc & 0x02) ? 1 : 0;
      break;
   default:
      ret = -1;
      break;
   }
   return ret;
}

/* ----------------------------- Linux/Unix API -----------------
 * mode = 1: ScrollLock
 *        2: CapsLock
 *        3: NumLock
 *        4: Shift
 * return = 0: not set, 1: set, -1: error -> 0
 */
int InkeyStatusApi(mode)
{
   int ret;
   
   ret = keyStateCons(mode);     // check console first
   if(ret == -1)
      ret = keyStateX11(mode);   // on error, check X11
   return ret == -1 ? 0 : ret; 
}
#endif

/* stand-alone test, compile with 
 *   cc inkeystatapi.c -lX11 ; a.out
 */
// int main()
// {
//    printf("Scroll   = %d\n", InkeyStatusApi(1));
//    printf("Capslock = %d\n", InkeyStatusApi(2));
//    printf("Numlock  = %d\n", InkeyStatusApi(3));
//    return 0;
// }
