/* Converts FoxPro/FoxBase/dBase .prg statements 
 *                  "DO /<anyPath>/<UdfName> [WITH ...]"
 *               or "DO C:\<anyPath>\<UdfName> [WITH ...]"
 * to proper syntax "DO <UdfName> [WITH ...]"
 * leaving the original statement as comment "*FOX* DO /<anyPath>/..."
 * Note: this cannot be done automatically by #command preprocessor.
 *
 * Note also, the <UdfName> is not an executable but a procedure name, 
 * where the source file (or object) containing it needs to be compiled 
 * or linked with your application, or referred by SET PROCEDURE TO ...
 * To run an external executable, use the RUN command instead.
 * 
 * The original file x.prg is renamed to x.prg.orig if not exists yet,
 * the converted file name is then the original file name x.prg.
 *
 * Compile Linux: FlagShip -io=b foxconvert.prg -o foxconvert
 * Execute Linux: ./foxconvert *.prg
 *
 * Compile Windows: FlagShip -io=b foxconvert.prg 
 * Execute Windows: foxconvert *.prg
 *
 * Then compile your application by: FlagShip -fox mymain.prg
 * or: FlagShip -fox *.prg -m -Mmymain
 */
 
#ifdef FS_WIN32
# define CRLF chr(13)+chr(10)          /* DOS/Windows end-of-line */
#else
# define CRLF chr(10)                  /* Unix/Linux end-of-line */
#endif

nFiles := pcount()                     // variable param count
if nFiles == 0 .or. ("," + lower(param(1)) + ",") $ ",/?,/h,/help,-h,--help,"
   ? "converts Fox syntax DO path/UDF [WITH ...] to proper FS syntax"
   ? "invoke:", exename(), "file1.prg [file2.prg file*.prg]"
   sleep(5)
   quit
endif

for ii := 1 to nFiles
   cFile := param(ii)                  // parameter 1..n = file name
   aFiles := {}
#ifdef FS_WIN32
   /* Windows does not expand wildcard at command-line, so
    * we do it manually. Linux/Unix passes expanded arguments.
    */
   if "?" $ cFile .or. "*" $ cFile
      aFiles := directory(cFile)
   endif
#endif
   if len(aFiles) > 0
      for iArr := 1 to len(aFiles)
         convert(aFiles[iArr, 1])      // expanded file name
      next
   else
      convert(cFile)                   // passed file name
   endif
next
// wait "done..."
quit

******************************************************************
FUNCTION convert(cFile)   
local fdIn, fdOut, iChanged, iReadErr, lRenamed 
local iPos1, iPos2, iPos3, iLine, cLine, cUdf
        
if lower(cFile) == "foxconvert.prg"          // not this file
   return
endif

? "Processing file '" + cFile + "' ... " 
if !file(cFile)
   ? "-- file '" + cFile + "' not available, ignored"
// inkey(5)
   return
endif
if !(".prg" $ lower(cFile)) .and. !(".fmt" $ lower(cFile))
   ? "-- file '" + cFile + "' has not source extension, ignored"
// inkey(5)
   return
endif

if !file(cFile + ".orig")
   rename (cFile) to (cFile + ".orig")       // only first time
   lRenamed := .T.
else
   lRenamed := .F.
endif

fdIn := fopen(cFile + ".orig", 0)            // open input file
if fdIn <= 0
   ? "-- cannot open file '" + cFile + ".orig', ignored"
   if lRenamed .and. file(cFile + ".orig") .and. !file(cFile)
      rename (cFile + ".orig") to (cFile)
   endif
// inkey(5)
   return
endif

fdOut := fcreate(cFile, 0)                   // create/open output file
if fdOut <= 0
   ? "-- cannot create file '" + cFile + "' from original '" + cFile + ".orig'"
// inkey(5)
   fclose(fdIn)  
   return
endif

iChanged := 0
iLine    := 0
iReadErr := 0
while iReadErr == 0
   cLine := freadtxt(fdIn)                   // read text line
   iReadErr := ferror()
   if iReadErr == -1 .and. empty(cLine)      // EOF detected
      exit
   endif
   iLine++                                   // line count (not displayed)
   cLine := strtran(cLine, chr(9), " ")      // TAB -> space
   
   if left(upper(ltrim(cLine)),3) == "DO "   // check for DO ... command
      iPos1 := at("DO ",upper(cLine))        
      for iPos2 := iPos1 + 3 to len(cLine)   // iPos2 = pos first UDF char
         if strpeek(cLine, iPos2) != 32      // skip spaces betw DO and UDF
            exit
         endif
      next
      for iPos3 := iPos2 + 1 to len(cLine)   // iPos3 = pos last UDF char
         if strpeek(cLine, iPos3) == 32      // test upto space or EOL
            iPos3 --
            exit
         endif
      next
      cUdf := substr(cLine, iPos2, iPos3 +1 - iPos2)  // UDF name (w/ path?)
      if !empty(cUdf) .and. ("/" $ cUdf .or. "\" $ cUdf)
         if "/" $ cUdf 
            cUdf := substr(cUdf, rat("/", cUdf) +1)   // extracted UDF name
         elseif "\" $ cUdf
            cUdf := substr(cUdf, rat("\", cUdf) +1)
         endif
         fwrite(fdOut, "*FOX* " + cLine + CRLF) // write original as comment
         cLine := left(cLine, iPos2 -1) + cUdf + substr(cLine, iPos3 +1) // fixed line
         iChanged++
      endif
   endif
   
   fwrite(fdOut, cLine + CRLF)      // write line back to file
   if ferror() != 0
      ? "-- write error to file '" + cFile + "' - aborted, use " + cFile + ".orig"
      exit
   endif
enddo
?? ltrim(iChanged), "changes"

fclose(fdIn)                        // close input
fclose(fdOut)                       // close output
return

* eof *
